@@ -47,37 +47,12 @@ unbind_resource (struct wl_resource *resource)
   wl_list_remove (wl_resource_get_link (resource));
 }
 
-static void
-transform_stage_point_fixed (MetaWaylandSurface *surface,
-                             wl_fixed_t x,
-                             wl_fixed_t y,
-                             wl_fixed_t *sx,
-                             wl_fixed_t *sy)
-{
-  float xf = 0.0f, yf = 0.0f;
-
-  if (surface->window)
-    {
-      ClutterActor *actor =
-        CLUTTER_ACTOR (meta_window_get_compositor_private (surface->window));
-
-      if (actor)
-        clutter_actor_transform_stage_point (actor,
-                                             wl_fixed_to_double (x),
-                                             wl_fixed_to_double (y),
-                                             &xf, &yf);
-    }
-
-  *sx = wl_fixed_from_double (xf);
-  *sy = wl_fixed_from_double (yf);
-}
-
 static void
 pointer_unmap_sprite (MetaWaylandSeat *seat)
 {
   if (seat->cursor_tracker)
     {
-      meta_cursor_tracker_set_sprite (seat->cursor_tracker,
+      meta_cursor_tracker_set_buffer (seat->cursor_tracker,
 				      NULL, 0, 0);
 
       if (seat->current_stage)
@@ -95,29 +70,20 @@ pointer_unmap_sprite (MetaWaylandSeat *seat)
 void
 meta_wayland_seat_update_sprite (MetaWaylandSeat *seat)
 {
-  ClutterBackend *backend;
-  CoglContext *context;
   struct wl_resource *buffer;
-  CoglTexture2D *texture;
 
   if (seat->cursor_tracker == NULL)
     return;
 
-  backend = clutter_get_default_backend ();
-  context = clutter_backend_get_cogl_context (backend);
   buffer = seat->sprite->buffer_ref.buffer->resource;
-  texture = cogl_wayland_texture_2d_new_from_buffer (context, buffer, NULL);
-
-  meta_cursor_tracker_set_sprite (seat->cursor_tracker,
-				  texture,
+  meta_cursor_tracker_set_buffer (seat->cursor_tracker,
+				  buffer,
 				  seat->hotspot_x,
 				  seat->hotspot_y);
 
   if (seat->current_stage)
     meta_cursor_tracker_queue_redraw (seat->cursor_tracker,
 				      CLUTTER_ACTOR (seat->current_stage));
-
-  cogl_object_unref (texture);
 }
 
 static void
@@ -175,26 +141,14 @@ seat_get_pointer (struct wl_client *client,
   MetaWaylandSeat *seat = wl_resource_get_user_data (resource);
   struct wl_resource *cr;
 
-  cr = wl_client_add_object (client, &wl_pointer_interface,
-                             &pointer_interface, id, seat);
+  cr = wl_resource_create (client, &wl_pointer_interface,
+			   MIN (META_WL_POINTER_VERSION, wl_resource_get_version (resource)), id);
+  wl_resource_set_implementation (cr, &pointer_interface, seat, unbind_resource);
   wl_list_insert (&seat->pointer.resource_list, wl_resource_get_link (cr));
-  wl_resource_set_destructor (cr, unbind_resource);
 
   if (seat->pointer.focus &&
       wl_resource_get_client (seat->pointer.focus->resource) == client)
-    {
-      MetaWaylandSurface *surface;
-      wl_fixed_t sx, sy;
-
-      surface = (MetaWaylandSurface *) seat->pointer.focus;
-      transform_stage_point_fixed (surface,
-                                   seat->pointer.x,
-                                   seat->pointer.y,
-                                   &sx, &sy);
-      meta_wayland_pointer_set_focus (&seat->pointer,
-                                  seat->pointer.focus,
-                                  sx, sy);
-    }
+    meta_wayland_pointer_set_focus (&seat->pointer, seat->pointer.focus);
 }
 
 static void
@@ -205,9 +159,10 @@ seat_get_keyboard (struct wl_client *client,
   MetaWaylandSeat *seat = wl_resource_get_user_data (resource);
   struct wl_resource *cr;
 
-  cr = wl_client_add_object (client, &wl_keyboard_interface, NULL, id, seat);
+  cr = wl_resource_create (client, &wl_keyboard_interface,
+			   MIN (META_WL_KEYBOARD_VERSION, wl_resource_get_version (resource)), id);
+  wl_resource_set_implementation (cr, NULL, seat, unbind_resource);
   wl_list_insert (&seat->keyboard.resource_list, wl_resource_get_link (cr));
-  wl_resource_set_destructor (cr, unbind_resource);
 
   wl_keyboard_send_keymap (cr,
                            WL_KEYBOARD_KEYMAP_FORMAT_XKB_V1,
@@ -217,8 +172,7 @@ seat_get_keyboard (struct wl_client *client,
   if (seat->keyboard.focus &&
       wl_resource_get_client (seat->keyboard.focus->resource) == client)
     {
-      meta_wayland_keyboard_set_focus (&seat->keyboard,
-                                   seat->keyboard.focus);
+      meta_wayland_keyboard_set_focus (&seat->keyboard, seat->keyboard.focus);
       meta_wayland_data_device_set_keyboard_focus (seat);
     }
 }
@@ -248,17 +202,17 @@ bind_seat (struct wl_client *client,
   MetaWaylandSeat *seat = data;
   struct wl_resource *resource;
 
-  resource = wl_client_add_object (client,
-                                   &wl_seat_interface,
-                                   &seat_interface,
-                                   id,
-                                   data);
+  resource = wl_resource_create (client, &wl_seat_interface,
+				 MIN (META_WL_SEAT_VERSION, version), id);
+  wl_resource_set_implementation (resource, &seat_interface, seat, unbind_resource);
   wl_list_insert (&seat->base_resource_list, wl_resource_get_link (resource));
-  wl_resource_set_destructor (resource, unbind_resource);
 
   wl_seat_send_capabilities (resource,
                              WL_SEAT_CAPABILITY_POINTER |
                              WL_SEAT_CAPABILITY_KEYBOARD);
+
+  if (version >= META_WL_SEAT_HAS_NAME)
+    wl_seat_send_name (resource, "seat0");
 }
 
 static void
@@ -276,15 +230,11 @@ meta_wayland_seat_new (struct wl_display *display,
 {
   MetaWaylandSeat *seat = g_new0 (MetaWaylandSeat, 1);
 
-  wl_signal_init (&seat->destroy_signal);
-
   seat->selection_data_source = NULL;
   wl_list_init (&seat->base_resource_list);
-  wl_signal_init (&seat->selection_signal);
-  wl_list_init (&seat->drag_resource_list);
-  wl_signal_init (&seat->drag_icon_signal);
+  wl_list_init (&seat->data_device_resource_list);
 
-  meta_wayland_pointer_init (&seat->pointer);
+  meta_wayland_pointer_init (&seat->pointer, is_native);
 
   meta_wayland_keyboard_init (&seat->keyboard, display, is_native);
 
@@ -297,13 +247,13 @@ meta_wayland_seat_new (struct wl_display *display,
   seat->hotspot_x = 16;
   seat->hotspot_y = 16;
 
-  wl_display_add_global (display, &wl_seat_interface, seat, bind_seat);
+  wl_global_create (display, &wl_seat_interface, META_WL_SEAT_VERSION, seat, bind_seat);
 
   return seat;
 }
 
 static void
-notify_motion (MetaWaylandSeat *seat,
+notify_motion (MetaWaylandSeat    *seat,
                const ClutterEvent *event)
 {
   MetaWaylandPointer *pointer = &seat->pointer;
@@ -313,91 +263,64 @@ notify_motion (MetaWaylandSeat *seat,
   pointer->x = wl_fixed_from_double (x);
   pointer->y = wl_fixed_from_double (y);
 
-  meta_wayland_seat_repick (seat,
-                        clutter_event_get_time (event),
-                        clutter_event_get_source (event));
+  meta_wayland_seat_repick (seat, event);
 
-  pointer->grab->interface->motion (pointer->grab,
-                                    clutter_event_get_time (event),
-                                    pointer->grab->x,
-                                    pointer->grab->y);
+  pointer->grab->interface->motion (pointer->grab, event);
 }
 
 static void
-handle_motion_event (MetaWaylandSeat *seat,
-                     const ClutterMotionEvent *event)
+handle_motion_event (MetaWaylandSeat    *seat,
+                     const ClutterEvent *event)
 {
-  notify_motion (seat, (const ClutterEvent *) event);
+  notify_motion (seat, event);
 }
 
 static void
-handle_button_event (MetaWaylandSeat *seat,
-                     const ClutterButtonEvent *event)
+handle_button_event (MetaWaylandSeat    *seat,
+                     const ClutterEvent *event)
 {
   MetaWaylandPointer *pointer = &seat->pointer;
   gboolean state = event->type == CLUTTER_BUTTON_PRESS;
   uint32_t button;
+  MetaWaylandSurface *surface;
 
-  notify_motion (seat, (const ClutterEvent *) event);
+  notify_motion (seat, event);
 
-  switch (event->button)
+  if (state && pointer->button_count == 1)
     {
-      /* The evdev input right and middle button numbers are swapped
-         relative to how Clutter numbers them */
-    case 2:
-      button = BTN_MIDDLE;
-      break;
+      button = clutter_event_get_button (event);
+      pointer->grab_button = button;
+      pointer->grab_time = clutter_event_get_time (event);
+      pointer->grab_x = pointer->x;
+      pointer->grab_y = pointer->y;
 
-    case 3:
-      button = BTN_RIGHT;
-      break;
-
-    default:
-      button = event->button + BTN_LEFT - 1;
-      break;
+      /* FIXME: synth a XI2 event and handle in display.c */
+      surface = pointer->current;
+      if (button == CLUTTER_BUTTON_PRIMARY &&
+	  surface &&
+	  surface->window &&
+	  surface->window->client_type == META_WINDOW_CLIENT_TYPE_WAYLAND)
+	{
+	  meta_window_raise (surface->window);
+	}
     }
 
-  if (state)
-    {
-      if (pointer->button_count == 0)
-        {
-          MetaWaylandSurface *surface = pointer->current;
-
-          pointer->grab_button = button;
-          pointer->grab_time = event->time;
-          pointer->grab_x = pointer->x;
-          pointer->grab_y = pointer->y;
-
-          if (button == BTN_LEFT &&
-              surface &&
-              surface->window &&
-              surface->window->client_type == META_WINDOW_CLIENT_TYPE_WAYLAND)
-            {
-              meta_window_raise (surface->window);
-            }
-        }
-
-      pointer->button_count++;
-    }
-  else
-    pointer->button_count--;
-
-  pointer->grab->interface->button (pointer->grab, event->time, button, state);
+  pointer->grab->interface->button (pointer->grab, event);
 
   if (pointer->button_count == 1)
     pointer->grab_serial = wl_display_get_serial (seat->display);
 }
 
 static void
-handle_scroll_event (MetaWaylandSeat *seat,
-                     const ClutterScrollEvent *event)
+handle_scroll_event (MetaWaylandSeat    *seat,
+                     const ClutterEvent *event)
 {
   enum wl_pointer_axis axis;
   wl_fixed_t value;
 
-  notify_motion (seat, (const ClutterEvent *) event);
+  notify_motion (seat, event);
 
-  switch (event->direction)
+  switch (clutter_event_get_scroll_direction (event))
     {
     case CLUTTER_SCROLL_UP:
       axis = WL_POINTER_AXIS_VERTICAL_SCROLL;
@@ -425,41 +348,64 @@ handle_scroll_event (MetaWaylandSeat *seat,
 
   if (seat->pointer.focus_resource)
     wl_pointer_send_axis (seat->pointer.focus_resource,
-                          event->time,
+                          clutter_event_get_time (event),
                           axis,
                           value);
 }
 
-void
+static int
+count_buttons (const ClutterEvent *event)
+{
+  static gint maskmap[5] =
+    {
+      CLUTTER_BUTTON1_MASK, CLUTTER_BUTTON2_MASK, CLUTTER_BUTTON3_MASK,
+      CLUTTER_BUTTON4_MASK, CLUTTER_BUTTON5_MASK
+    };
+  ClutterModifierType mod_mask;
+  int i, count;
+
+  mod_mask = clutter_event_get_state (event);
+  count = 0;
+  for (i = 0; i < 5; i++)
+    {
+      if (mod_mask & maskmap[i])
+	count++;
+    }
+
+  return count;
+}
+
+gboolean
 meta_wayland_seat_handle_event (MetaWaylandSeat *seat,
                                 const ClutterEvent *event)
 {
+  seat->pointer.button_count = count_buttons (event);
+
   switch (event->type)
     {
     case CLUTTER_MOTION:
-      handle_motion_event (seat,
-                           (const ClutterMotionEvent *) event);
+      handle_motion_event (seat, event);
       break;
 
     case CLUTTER_BUTTON_PRESS:
     case CLUTTER_BUTTON_RELEASE:
-      handle_button_event (seat,
-                           (const ClutterButtonEvent *) event);
+      handle_button_event (seat, event);
       break;
 
     case CLUTTER_KEY_PRESS:
     case CLUTTER_KEY_RELEASE:
-      meta_wayland_keyboard_handle_event (&seat->keyboard,
-                                          (const ClutterKeyEvent *) event);
-      break;
+      return meta_wayland_keyboard_handle_event (&seat->keyboard,
+                                                 (const ClutterKeyEvent *) event);
 
     case CLUTTER_SCROLL:
-      handle_scroll_event (seat, (const ClutterScrollEvent *) event);
+      handle_scroll_event (seat, event);
       break;
 
     default:
       break;
     }
+
+  return FALSE;
 }
 
 static void
@@ -482,14 +428,18 @@ update_pointer_position_for_actor (MetaWaylandPointer *pointer,
    case Clutter will have already performed a pick so we can avoid
    redundantly doing another one */
 void
-meta_wayland_seat_repick (MetaWaylandSeat *seat,
-                          uint32_t time,
-                          ClutterActor *actor)
+meta_wayland_seat_repick (MetaWaylandSeat    *seat,
+			  const ClutterEvent *for_event)
 {
+  ClutterActor       *actor   = NULL;
   MetaWaylandPointer *pointer = &seat->pointer;
   MetaWaylandSurface *surface = NULL;
 
-  if (actor == NULL && seat->current_stage)
+  if (for_event)
+    {
+      actor = clutter_event_get_source (for_event);
+    }
+  else if (seat->current_stage)
     {
       ClutterStage *stage = CLUTTER_STAGE (seat->current_stage);
       actor = clutter_stage_get_actor_at_pos (stage,
@@ -527,16 +477,8 @@ meta_wayland_seat_repick (MetaWaylandSeat *seat,
       const MetaWaylandPointerGrabInterface *interface =
         pointer->grab->interface;
       interface->focus (pointer->grab,
-                        surface,
-                        pointer->current_x, pointer->current_y);
+                        surface, for_event);
     }
-
-  if (pointer->grab->focus)
-    transform_stage_point_fixed (pointer->grab->focus,
-                                 pointer->x,
-                                 pointer->y,
-                                 &pointer->grab->x,
-                                 &pointer->grab->y);
 }
 
 void
@@ -547,7 +489,5 @@ meta_wayland_seat_free (MetaWaylandSeat *seat)
   meta_wayland_pointer_release (&seat->pointer);
   meta_wayland_keyboard_release (&seat->keyboard);
 
-  wl_signal_emit (&seat->destroy_signal, seat);
-
   g_slice_free (MetaWaylandSeat, seat);
 }
