@@ -25,6 +25,7 @@
 #include <clutter/clutter.h>
 #include <clutter/wayland/clutter-wayland-compositor.h>
 #include <clutter/wayland/clutter-wayland-surface.h>
+#include <cogl/cogl-wayland-server.h>
 
 #include <glib.h>
 #include <sys/time.h>
@@ -64,8 +65,9 @@ static void
 surface_process_damage (MetaWaylandSurface *surface,
                         cairo_region_t *region)
 {
-  if (surface->window &&
-      surface->buffer_ref.buffer)
+  g_assert (surface->window);
+
+  if (surface->buffer_ref.buffer)
     {
       MetaWindowActor *window_actor =
         META_WINDOW_ACTOR (meta_window_get_compositor_private (surface->window));
@@ -111,7 +113,7 @@ static void
 meta_wayland_surface_attach (struct wl_client *wayland_client,
                              struct wl_resource *wayland_surface_resource,
                              struct wl_resource *wayland_buffer_resource,
-                             gint32 sx, gint32 sy)
+                             gint32 dx, gint32 dy)
 {
   MetaWaylandSurface *surface =
     wl_resource_get_user_data (wayland_surface_resource);
@@ -130,8 +132,8 @@ meta_wayland_surface_attach (struct wl_client *wayland_client,
   if (surface->pending.buffer)
     wl_list_remove (&surface->pending.buffer_destroy_listener.link);
 
-  surface->pending.sx = sx;
-  surface->pending.sy = sy;
+  surface->pending.dx = dx;
+  surface->pending.dy = dy;
   surface->pending.buffer = buffer;
   surface->pending.newly_attached = TRUE;
 
@@ -247,36 +249,86 @@ meta_wayland_surface_commit (struct wl_client *client,
   if (surface->pending.newly_attached &&
       surface->buffer_ref.buffer != surface->pending.buffer)
     {
+      MetaWaylandBuffer *buffer = surface->pending.buffer;
+      CoglContext *ctx =
+        clutter_backend_get_cogl_context (clutter_get_default_backend ());
+      CoglError *catch_error = NULL;
+      CoglTexture *texture =
+        COGL_TEXTURE (cogl_wayland_texture_2d_new_from_buffer (ctx,
+                                                               buffer->resource,
+                                                               &catch_error));
+      if (!texture)
+        {
+          cogl_error_free (catch_error);
+	  meta_warning ("Could not import pending buffer, ignoring commit\n");
+	  return;
+        }
+      else
+        {
+	  buffer->texture = texture;
+          buffer->width = cogl_texture_get_width (texture);
+          buffer->height = cogl_texture_get_height (texture);
+        }
+
       /* Note: we set this before informing any window-actor since the
        * window actor will expect to find the new buffer within the
        * surface. */
       meta_wayland_buffer_reference (&surface->buffer_ref,
                                      surface->pending.buffer);
+    }
+
+  if (!surface->buffer_ref.buffer)
+    {
+      meta_warning ("Commit without a buffer? Ignoring\n");
+      return;
+    }
+
+  if (surface->initial_state)
+    {
+      MetaDisplay *display = meta_get_display ();
+      int width;
+      int height;
+
+      width = surface->buffer_ref.buffer->width;
+      height = surface->buffer_ref.buffer->height;
+
+      /* This will free and clear initial_state */
+      surface->window = meta_window_new_for_wayland (display, width, height, surface);
+    }
+
+  if (surface == compositor->seat->sprite)
+    meta_wayland_seat_update_sprite (compositor->seat);
+  else if (surface->window)
+    {
+      MetaWindow *window = surface->window;
 
       if (surface->pending.buffer)
-        {
-          MetaWaylandBuffer *buffer = surface->pending.buffer;
+	{
+	  MetaWaylandBuffer *buffer = surface->pending.buffer;
+	  MetaWindowActor *window_actor = META_WINDOW_ACTOR (meta_window_get_compositor_private (window));
 
-          if (surface->window)
-            {
-              MetaWindow *window = surface->window;
-              MetaWindowActor *window_actor =
-                META_WINDOW_ACTOR (meta_window_get_compositor_private (window));
-              MetaRectangle rect;
+	  meta_window_actor_attach_wayland_buffer (window_actor, buffer);
+	}
 
-              meta_window_get_input_rect (surface->window, &rect);
+      /* We resize X based surfaces according to X events */
+      if (window->client_type == META_WINDOW_CLIENT_TYPE_WAYLAND)
+	{
+	  int new_width;
+	  int new_height;
 
-              if (window_actor)
-                meta_window_actor_attach_wayland_buffer (window_actor, buffer);
+	  new_width = surface->buffer_ref.buffer->width;
+	  new_height = surface->buffer_ref.buffer->height;
+	  if (new_width != window->rect.width ||
+	      new_height != window->rect.height ||
+	      surface->pending.dx != 0 ||
+	      surface->pending.dy != 0)
+	    meta_window_move_resize_wayland (surface->window, new_width, new_height,
+					     surface->pending.dx, surface->pending.dy);
+	}
 
-              /* XXX: we resize X based surfaces according to X events */
-              if (surface->xid == 0 &&
-                  (buffer->width != rect.width || buffer->height != rect.height))
-                meta_window_resize (surface->window, FALSE, buffer->width, buffer->height);
-            }
-          else if (surface == compositor->seat->sprite)
-            meta_wayland_seat_update_sprite (compositor->seat);
-        }
+      meta_window_set_opaque_region (surface->window, surface->pending.opaque_region);
+      meta_window_set_input_region (surface->window, surface->pending.input_region);
+      surface_process_damage (surface, surface->pending.damage);
     }
 
   if (surface->pending.buffer)
@@ -284,20 +336,11 @@ meta_wayland_surface_commit (struct wl_client *client,
       wl_list_remove (&surface->pending.buffer_destroy_listener.link);
       surface->pending.buffer = NULL;
     }
-  surface->pending.sx = 0;
-  surface->pending.sy = 0;
+  surface->pending.dx = 0;
+  surface->pending.dy = 0;
   surface->pending.newly_attached = FALSE;
-
-  if (surface->window)
-    {
-      meta_window_set_opaque_region (surface->window, surface->pending.opaque_region);
-      g_clear_pointer (&surface->pending.opaque_region, cairo_region_destroy);
-
-      meta_window_set_input_region (surface->window, surface->pending.input_region);
-      g_clear_pointer (&surface->pending.input_region, cairo_region_destroy);
-    }
-
-  surface_process_damage (surface, surface->pending.damage);
+  g_clear_pointer (&surface->pending.opaque_region, cairo_region_destroy);
+  g_clear_pointer (&surface->pending.input_region, cairo_region_destroy);
   empty_region (surface->pending.damage);
 
   /* wl_surface.frame */
@@ -364,7 +407,6 @@ meta_wayland_surface_free (MetaWaylandSurface *surface)
       meta_wayland_pointer_destroy_focus (&compositor->seat->pointer);
 
       g_assert (surface != compositor->seat->pointer.focus);
-      g_assert (surface != compositor->seat->pointer.grab->focus);
     }
 
  if (compositor->implicit_grab_surface == surface)
@@ -450,58 +492,118 @@ shell_surface_pong (struct wl_client *client,
 {
 }
 
-typedef struct _MetaWaylandGrab
+typedef struct
 {
-  MetaWaylandPointerGrab grab;
-  MetaWaylandSurfaceExtension *shell_surface;
-  struct wl_listener shell_surface_destroy_listener;
-  MetaWaylandPointer *pointer;
-} MetaWaylandGrab;
-
-typedef struct _MetaWaylandMoveGrab
-{
-  MetaWaylandGrab base;
-  wl_fixed_t dx, dy;
+  MetaWaylandPointerGrab  generic;
+  MetaWaylandSurface     *surface;
+  struct wl_listener      surface_destroy_listener;
+  wl_fixed_t              dx, dy;
 } MetaWaylandMoveGrab;
 
 static void
-destroy_shell_surface_grab_listener (struct wl_listener *listener,
-                                     void *data)
+end_move_grab (MetaWaylandMoveGrab *move_grab)
 {
-  MetaWaylandGrab *grab = wl_container_of (listener, grab,
-                                           shell_surface_destroy_listener);
-  grab->shell_surface = NULL;
+  if (move_grab->surface)
+    {
+      move_grab->surface = NULL;
+      wl_list_remove (&move_grab->surface_destroy_listener.link);
+    }
 
-  /* XXX: Could we perhaps just stop the grab here so we don't have
-   * to consider grab->shell_surface becoming NULL in grab interface
-   * callbacks? */
+  meta_wayland_pointer_end_grab (move_grab->generic.pointer);
+  g_slice_free (MetaWaylandMoveGrab, move_grab);
 }
 
-typedef enum _GrabCursor
+static void
+move_grab_lose_surface (struct wl_listener *listener, void *data)
 {
-  GRAB_CURSOR_MOVE,
-} GrabCursor;
+  MetaWaylandMoveGrab *move_grab =
+    wl_container_of (listener, move_grab, surface_destroy_listener);
+
+  move_grab->surface = NULL;
+  end_move_grab (move_grab);
+}
 
 static void
-grab_pointer (MetaWaylandGrab *grab,
-              const MetaWaylandPointerGrabInterface *interface,
-              MetaWaylandSurfaceExtension *shell_surface,
-              MetaWaylandPointer *pointer,
-              GrabCursor cursor)
+move_grab_focus (MetaWaylandPointerGrab *grab,
+                 MetaWaylandSurface     *surface,
+		 const ClutterEvent     *event)
 {
-  /* TODO: popup_grab_end (pointer); */
+}
 
-  grab->grab.interface = interface;
-  grab->shell_surface = shell_surface;
-  grab->shell_surface_destroy_listener.notify =
-    destroy_shell_surface_grab_listener;
-  wl_resource_add_destroy_listener (shell_surface->resource,
-                                    &grab->shell_surface_destroy_listener);
+static void
+move_grab_motion (MetaWaylandPointerGrab *grab,
+		  const ClutterEvent     *event)
+{
+  MetaWaylandMoveGrab *move_grab = (MetaWaylandMoveGrab *)grab;
+  MetaWaylandPointer *pointer = grab->pointer;
 
-  grab->pointer = pointer;
-  grab->grab.focus = shell_surface->surface;
+  meta_window_move (move_grab->surface->window,
+                    TRUE,
+                    wl_fixed_to_int (pointer->x + move_grab->dx),
+                    wl_fixed_to_int (pointer->y + move_grab->dy));
+}
 
-  meta_wayland_pointer_start_grab (pointer, &grab->grab);
+static void
+move_grab_button (MetaWaylandPointerGrab *grab,
+		  const ClutterEvent     *event)
+{
+  MetaWaylandMoveGrab *move_grab = (MetaWaylandMoveGrab *)grab;
+  MetaWaylandPointer *pointer = grab->pointer;
+
+  if (pointer->button_count == 0 &&
+      clutter_event_type (event) == CLUTTER_BUTTON_RELEASE)
+    end_move_grab (move_grab);
+}
+
+static const MetaWaylandPointerGrabInterface move_grab_interface = {
+    move_grab_focus,
+    move_grab_motion,
+    move_grab_button,
+};
+
+static void
+shell_surface_move (struct wl_client *client,
+                    struct wl_resource *resource,
+                    struct wl_resource *seat_resource,
+                    guint32 serial)
+{
+  MetaWaylandSeat *seat = wl_resource_get_user_data (seat_resource);
+  MetaWaylandSurfaceExtension *shell_surface = wl_resource_get_user_data (resource);
+  MetaWindow *window;
+  MetaWaylandMoveGrab *move_grab;
+  MetaRectangle rect;
+
+  if (seat->pointer.button_count == 0 ||
+      seat->pointer.grab_serial != serial ||
+      seat->pointer.focus != shell_surface->surface)
+    return;
+
+  window = shell_surface->surface->window;
+  if (!window)
+    return;
+
+  /* small hack, this should be handled by window->shaken_loose when we
+     move everything to display.c */
+  if (window->fullscreen)
+    meta_window_unmake_fullscreen (window);
+
+  move_grab = g_slice_new (MetaWaylandMoveGrab);
+
+  meta_window_get_input_rect (shell_surface->surface->window,
+                              &rect);
+
+  move_grab->generic.interface = &move_grab_interface;
+  move_grab->generic.pointer = &seat->pointer;
+
+  move_grab->surface = shell_surface->surface;
+  move_grab->surface_destroy_listener.notify = move_grab_lose_surface;
+  wl_resource_add_destroy_listener (shell_surface->surface->resource,
+				    &move_grab->surface_destroy_listener);
+
+  move_grab->dx = wl_fixed_from_int (rect.x) - seat->pointer.grab_x;
+  move_grab->dy = wl_fixed_from_int (rect.y) - seat->pointer.grab_y;
+
+  meta_wayland_pointer_start_grab (&seat->pointer, (MetaWaylandPointerGrab*)move_grab);
 
   /* TODO: send_grab_cursor (cursor); */
 
@@ -514,112 +616,7 @@ grab_pointer (MetaWaylandGrab *grab,
    * XXX: For now we just focus the surface directly associated with
    * the grab.
    */
-  meta_wayland_pointer_set_focus (pointer,
-                                  grab->shell_surface->surface,
-                                  wl_fixed_from_int (0),
-                                  wl_fixed_from_int (0));
-}
-
-static void
-release_pointer (MetaWaylandGrab *grab)
-{
-  if (grab->shell_surface)
-    wl_list_remove (&grab->shell_surface_destroy_listener.link);
-
-  meta_wayland_pointer_end_grab (grab->pointer);
-}
-
-static void
-noop_grab_focus (MetaWaylandPointerGrab *grab,
-                 MetaWaylandSurface *surface,
-                 wl_fixed_t x,
-                 wl_fixed_t y)
-{
-  grab->focus = NULL;
-}
-
-static void
-move_grab_motion (MetaWaylandPointerGrab *grab,
-                  uint32_t time,
-                  wl_fixed_t x,
-                  wl_fixed_t y)
-{
-  MetaWaylandMoveGrab *move = (MetaWaylandMoveGrab *)grab;
-  MetaWaylandPointer *pointer = move->base.pointer;
-  MetaWaylandSurfaceExtension *shell_surface = move->base.shell_surface;
-
-  if (!shell_surface)
-    return;
-
-  meta_window_move (shell_surface->surface->window,
-                    TRUE,
-                    wl_fixed_to_int (pointer->x + move->dx),
-                    wl_fixed_to_int (pointer->y + move->dy));
-}
-
-static void
-move_grab_button (MetaWaylandPointerGrab *pointer_grab,
-                  uint32_t time,
-                  uint32_t button,
-                  uint32_t state_w)
-{
-  MetaWaylandGrab *grab =
-    wl_container_of (pointer_grab, grab, grab);
-  MetaWaylandMoveGrab *move = (MetaWaylandMoveGrab *)grab;
-  MetaWaylandPointer *pointer = grab->pointer;
-  enum wl_pointer_button_state state = state_w;
-
-  if (pointer->button_count == 0 && state == WL_POINTER_BUTTON_STATE_RELEASED)
-    {
-      release_pointer (grab);
-      g_slice_free (MetaWaylandMoveGrab, move);
-    }
-}
-
-static const MetaWaylandPointerGrabInterface move_grab_interface = {
-    noop_grab_focus,
-    move_grab_motion,
-    move_grab_button,
-};
-
-static void
-start_surface_move (MetaWaylandSurfaceExtension *shell_surface,
-                    MetaWaylandSeat *seat)
-{
-  MetaWaylandMoveGrab *move;
-  MetaRectangle rect;
-
-  g_return_if_fail (shell_surface != NULL);
-
-  /* TODO: check if the surface is fullscreen when we support fullscreen */
-
-  move = g_slice_new (MetaWaylandMoveGrab);
-
-  meta_window_get_input_rect (shell_surface->surface->window,
-                              &rect);
-
-  move->dx = wl_fixed_from_int (rect.x) - seat->pointer.grab_x;
-  move->dy = wl_fixed_from_int (rect.y) - seat->pointer.grab_y;
-
-  grab_pointer (&move->base, &move_grab_interface, shell_surface,
-                &seat->pointer, GRAB_CURSOR_MOVE);
-}
-
-static void
-shell_surface_move (struct wl_client *client,
-                    struct wl_resource *resource,
-                    struct wl_resource *seat_resource,
-                    guint32 serial)
-{
-  MetaWaylandSeat *seat = wl_resource_get_user_data (seat_resource);
-  MetaWaylandSurfaceExtension *shell_surface = wl_resource_get_user_data (resource);
-
-  if (seat->pointer.button_count == 0 ||
-      seat->pointer.grab_serial != serial ||
-      seat->pointer.focus != shell_surface->surface)
-    return;
-
-  start_surface_move (shell_surface, seat);
+  meta_wayland_pointer_set_focus (&seat->pointer, shell_surface->surface);
 }
 
 static void
@@ -632,34 +629,6 @@ shell_surface_resize (struct wl_client *client,
   g_warning ("TODO: support shell_surface_resize request");
 }
 
-static void
-ensure_surface_window (MetaWaylandSurface *surface)
-{
-  MetaDisplay *display = meta_get_display ();
-
-  if (!surface->window)
-    {
-      int width, height;
-
-      if (surface->buffer_ref.buffer)
-        {
-          MetaWaylandBuffer *buffer = surface->buffer_ref.buffer;
-          width = buffer->width;
-          height = buffer->width;
-        }
-      else
-        {
-          width = 0;
-          height = 0;
-        }
-
-      surface->window =
-        meta_window_new_for_wayland (display, width, height, surface);
-
-      meta_window_calc_showing (surface->window);
-    }
-}
-
 static void
 shell_surface_set_toplevel (struct wl_client *client,
                             struct wl_resource *resource)
@@ -672,9 +641,19 @@ shell_surface_set_toplevel (struct wl_client *client,
   if (client == compositor->xwayland_client)
     return;
 
-  ensure_surface_window (surface);
+  if (surface->window)
+    {
+      if (surface->window->fullscreen)
+	meta_window_unmake_fullscreen (surface->window);
+      if (meta_window_get_maximized (surface->window) != 0)
+	meta_window_unmaximize (surface->window, META_MAXIMIZE_HORIZONTAL | META_MAXIMIZE_VERTICAL);
+    }
+  else
+    {
+      ensure_initial_state (surface);
 
-  meta_window_unmake_fullscreen (surface->window);
+      surface->initial_state->initial_type = META_WAYLAND_SURFACE_TOPLEVEL;
+    }
 }
 
 static void
@@ -687,13 +666,22 @@ shell_surface_set_transient (struct wl_client *client,
 {
   MetaWaylandSurfaceExtension *shell_surface = wl_resource_get_user_data (resource);
   MetaWaylandSurface *surface = shell_surface->surface;
+  MetaWaylandSurface *parent_surface = wl_resource_get_user_data (parent);
   MetaWaylandCompositor *compositor = surface->compositor;
 
   /* NB: Surfaces from xwayland become managed based on X events. */
   if (client == compositor->xwayland_client)
     return;
 
-  ensure_surface_window (surface);
+  if (surface->window)
+    meta_window_set_transient_for (surface->window, parent_surface->window);
+  else
+    {
+      ensure_initial_state (surface);
+
+      surface->initial_state->initial_type = META_WAYLAND_SURFACE_TOPLEVEL;
+      surface->initial_state->transient_for = parent;
+    }
 }
 
 static void
@@ -711,21 +699,51 @@ shell_surface_set_fullscreen (struct wl_client *client,
   if (client == compositor->xwayland_client)
     return;
 
-  ensure_surface_window (surface);
+  if (surface->window)
+    meta_window_make_fullscreen (surface->window);
+  else
+    {
+      ensure_initial_state (surface);
 
-  meta_window_make_fullscreen (surface->window);
+      surface->initial_state->initial_type = META_WAYLAND_SURFACE_FULLSCREEN;
+    }
 }
 
 static void
 shell_surface_set_popup (struct wl_client *client,
                          struct wl_resource *resource,
-                         struct wl_resource *seat,
+                         struct wl_resource *seat_resource,
                          guint32 serial,
                          struct wl_resource *parent,
                          gint32 x,
                          gint32 y,
                          guint32 flags)
 {
+  MetaWaylandSurfaceExtension *shell_surface = wl_resource_get_user_data (resource);
+  MetaWaylandSurface *surface = shell_surface->surface;
+  MetaWaylandCompositor *compositor = surface->compositor;
+  MetaWaylandSeat *seat = compositor->seat;
+
+  if (serial < seat->pointer.click_serial)
+    {
+      /* stale request */
+      return;
+    }
+
+  if (surface->window)
+    {
+      meta_warning ("Client set_popup() on an already visible window, this is not supported\n");
+    }
+  else
+    {
+      ensure_initial_state (surface);
+
+      surface->initial_state->initial_type = META_WAYLAND_SURFACE_POPUP;
+      surface->initial_state->transient_for = parent;
+      surface->initial_state->x = x;
+      surface->initial_state->y = y;
+    }
+
 }
 
 static void
@@ -733,7 +751,22 @@ shell_surface_set_maximized (struct wl_client *client,
                              struct wl_resource *resource,
                              struct wl_resource *output)
 {
-  g_warning ("TODO: support shell_surface_set_maximized request");
+  MetaWaylandSurfaceExtension *shell_surface = wl_resource_get_user_data (resource);
+  MetaWaylandSurface *surface = shell_surface->surface;
+  MetaWaylandCompositor *compositor = surface->compositor;
+
+  /* NB: Surfaces from xwayland become managed based on X events. */
+  if (client == compositor->xwayland_client)
+    return;
+
+  if (surface->window)
+    meta_window_maximize (surface->window, META_MAXIMIZE_HORIZONTAL | META_MAXIMIZE_VERTICAL);
+  else
+    {
+      ensure_initial_state (surface);
+
+      surface->initial_state->initial_type = META_WAYLAND_SURFACE_MAXIMIZED;
+    }
 }
 
 static void
@@ -813,10 +846,11 @@ destroy_surface_extension (struct wl_resource *resource)
   g_free (extension);
 }
 
-static void
+static MetaWaylandSurfaceExtension *
 create_surface_extension (struct wl_client          *client,
 			  struct wl_resource        *master_resource,
 			  guint32                    id,
+			  int                        max_version,
 			  MetaWaylandSurface        *surface,
 			  const struct wl_interface *interface,
 			  const void                *implementation)
@@ -826,7 +860,7 @@ create_surface_extension (struct wl_client          *client,
   extension = g_new0 (MetaWaylandSurfaceExtension, 1);
 
   extension->resource = wl_resource_create (client, interface,
-					    wl_resource_get_version (master_resource), id);
+					    MIN (max_version, wl_resource_get_version (master_resource)), id);
   wl_resource_set_implementation (extension->resource, implementation,
 				  extension, destroy_surface_extension);
 
@@ -834,6 +868,8 @@ create_surface_extension (struct wl_client          *client,
   extension->surface_destroy_listener.notify = extension_handle_surface_destroy;
   wl_resource_add_destroy_listener (surface->resource,
                                     &extension->surface_destroy_listener);
+
+  return extension;
 }
 
 static void
@@ -844,7 +880,7 @@ get_shell_surface (struct wl_client *client,
 {
   MetaWaylandSurface *surface = wl_resource_get_user_data (surface_resource);
 
-  if (surface->has_shell_surface)
+  if (surface->shell_surface)
     {
       wl_resource_post_error (surface_resource,
                               WL_DISPLAY_ERROR_INVALID_OBJECT,
@@ -852,10 +888,10 @@ get_shell_surface (struct wl_client *client,
       return;
     }
 
-  create_surface_extension (client, resource, id, surface,
-			    &wl_shell_surface_interface,
-			    &meta_wayland_shell_surface_interface);
-  surface->has_shell_surface = TRUE;
+  surface->shell_surface = create_surface_extension (client, resource, id,
+                                                     META_WL_SHELL_SURFACE_VERSION, surface,
+                                                     &wl_shell_surface_interface,
+                                                     &meta_wayland_shell_surface_interface);
 }
 
 static const struct wl_shell_interface meta_wayland_shell_interface =
@@ -871,7 +907,8 @@ bind_shell (struct wl_client *client,
 {
   struct wl_resource *resource;
 
-  resource = wl_resource_create (client, &wl_shell_interface, version, id);
+  resource = wl_resource_create (client, &wl_shell_interface,
+				 MIN (META_WL_SHELL_VERSION, version), id);
   wl_resource_set_implementation (resource, &meta_wayland_shell_interface, data, NULL);
 }
 
@@ -946,7 +983,7 @@ get_gtk_surface (struct wl_client *client,
 {
   MetaWaylandSurface *surface = wl_resource_get_user_data (surface_resource);
 
-  if (surface->has_gtk_surface)
+  if (surface->gtk_surface)
     {
       wl_resource_post_error (surface_resource,
                               WL_DISPLAY_ERROR_INVALID_OBJECT,
@@ -954,10 +991,10 @@ get_gtk_surface (struct wl_client *client,
       return;
     }
 
-  create_surface_extension (client, resource, id, surface,
-			    &gtk_surface_interface,
-			    &meta_wayland_gtk_surface_interface);
-  surface->has_gtk_surface = TRUE;
+  surface->gtk_surface = create_surface_extension (client, resource, id,
+                                                   META_GTK_SURFACE_VERSION, surface,
+                                                   &gtk_surface_interface,
+                                                   &meta_wayland_gtk_surface_interface);
 }
 
 static const struct gtk_shell_interface meta_wayland_gtk_shell_interface =
@@ -973,7 +1010,8 @@ bind_gtk_shell (struct wl_client *client,
 {
   struct wl_resource *resource;
 
-  resource = wl_resource_create (client, &gtk_shell_interface, version, id);
+  resource = wl_resource_create (client, &gtk_shell_interface,
+				 MIN (META_GTK_SHELL_VERSION, version), id);
   wl_resource_set_implementation (resource, &meta_wayland_gtk_shell_interface, data, NULL);
 
   /* FIXME: ask the plugin */
@@ -984,12 +1022,14 @@ void
 meta_wayland_init_shell (MetaWaylandCompositor *compositor)
 {
   if (wl_global_create (compositor->wayland_display,
-			&wl_shell_interface, 1,
+			&wl_shell_interface,
+			META_WL_SHELL_VERSION,
 			compositor, bind_shell) == NULL)
     g_error ("Failed to register a global shell object");
 
   if (wl_global_create (compositor->wayland_display,
-			&gtk_shell_interface, 1,
+			&gtk_shell_interface,
+			META_GTK_SHELL_VERSION,
 			compositor, bind_gtk_shell) == NULL)
     g_error ("Failed to register a global gtk-shell object");
 }
@@ -999,10 +1039,56 @@ meta_wayland_surface_set_initial_state (MetaWaylandSurface *surface,
 					MetaWindow         *window)
 {
   MetaWaylandSurfaceInitialState *initial = surface->initial_state;
+  MetaWaylandCompositor *compositor = surface->compositor;
+  MetaWaylandSeat *seat = compositor->seat;
 
   if (initial == NULL)
     return;
 
+  window->type = META_WINDOW_NORMAL;
+
+  /* Note that we poke at the bits directly here, because we're
+     in the middle of meta_window_new_shared() */
+  switch (initial->initial_type)
+    {
+    case META_WAYLAND_SURFACE_TOPLEVEL:
+      break;
+    case META_WAYLAND_SURFACE_FULLSCREEN:
+      window->fullscreen = TRUE;
+      break;
+    case META_WAYLAND_SURFACE_MAXIMIZED:
+      window->maximized_horizontally = window->maximized_vertically = TRUE;
+      break;
+    case META_WAYLAND_SURFACE_POPUP:
+      window->override_redirect = TRUE;
+      window->type = META_WINDOW_DROPDOWN_MENU;
+      window->mapped = TRUE;
+      window->showing_for_first_time = FALSE;
+      window->placed = TRUE;
+      if (!meta_wayland_pointer_start_popup_grab (&seat->pointer, surface))
+	wl_shell_surface_send_popup_done (surface->shell_surface->resource);
+      break;
+    default:
+      g_assert_not_reached ();
+    }
+
+  if (initial->transient_for)
+    {
+      MetaWaylandSurface *parent = wl_resource_get_user_data (initial->transient_for);
+      if (parent && parent->window)
+	{
+	  window->transient_for = g_object_ref (parent->window);
+
+	  if (initial->initial_type == META_WAYLAND_SURFACE_POPUP)
+	    {
+	      window->rect.x = parent->window->rect.x + initial->x;
+	      window->rect.y = parent->window->rect.y + initial->y;
+	    }
+	}
+
+      window->type = META_WINDOW_DIALOG;
+    }
+
   if (initial->title)
     meta_window_set_title (window, initial->title);
 
@@ -1017,6 +1103,8 @@ meta_wayland_surface_set_initial_state (MetaWaylandSurface *surface,
 				       initial->gtk_application_object_path,
 				       initial->gtk_window_object_path);
 
+  meta_window_type_changed (window);
+
   free_initial_state (initial);
   surface->initial_state = NULL;
 }
@@ -1045,3 +1133,14 @@ free_initial_state (MetaWaylandSurfaceInitialState *initial)
 
   g_slice_free (MetaWaylandSurfaceInitialState, initial);
 }
+
+void
+meta_wayland_surface_configure_notify (MetaWaylandSurface *surface,
+				       int                 new_width,
+				       int                 new_height,
+				       int                 edges)
+{
+  if (surface->shell_surface)
+    wl_shell_surface_send_configure (surface->shell_surface->resource,
+				     edges, new_width, new_height);
+}
