@@ -45,10 +45,6 @@ xserver_set_window_id (struct wl_client *client,
   MetaDisplay *display = meta_get_display ();
   MetaWindow *window;
 
-  g_return_if_fail (surface->xid == None);
-
-  surface->xid = xid;
-
   window  = meta_display_lookup_x_window (display, xid);
   if (window)
     {
@@ -87,8 +83,8 @@ bind_xserver (struct wl_client *client,
   if (client != compositor->xwayland_client)
     return;
 
-  compositor->xserver_resource =
-    wl_resource_create (client, &xserver_interface, version, id);
+  compositor->xserver_resource = wl_resource_create (client, &xserver_interface,
+						     MIN (META_XSERVER_VERSION, version), id);
   wl_resource_set_implementation (compositor->xserver_resource,
 				  &xserver_implementation, compositor, NULL);
 
@@ -269,7 +265,7 @@ bind_to_unix_socket (int display)
 }
 
 static void
-uncloexec_and_setpgid (gpointer user_data)
+uncloexec (gpointer user_data)
 {
   int fd = GPOINTER_TO_INT (user_data);
 
@@ -278,10 +274,6 @@ uncloexec_and_setpgid (gpointer user_data)
   int flags = fcntl (fd, F_GETFD);
   if (flags != -1)
     fcntl (fd, F_SETFD, flags & ~FD_CLOEXEC);
-
-  /* Put this process in a background process group, so that Ctrl-C
-     goes to mutter only */
-  setpgid (0, 0);
 }
 
 static void
@@ -323,7 +315,8 @@ meta_xwayland_start (MetaWaylandCompositor *compositor)
   GError *error;
 
   wl_global_create (compositor->wayland_display,
-		    &xserver_interface, 1,
+		    &xserver_interface,
+		    META_XSERVER_VERSION,
 		    compositor, bind_xserver);
 
   do
@@ -401,7 +394,7 @@ meta_xwayland_start (MetaWaylandCompositor *compositor)
                      G_SPAWN_DO_NOT_REAP_CHILD |
                      G_SPAWN_STDOUT_TO_DEV_NULL |
                      G_SPAWN_STDERR_TO_DEV_NULL,
-                     uncloexec_and_setpgid,
+                     uncloexec,
                      GINT_TO_POINTER (sp[1]),
                      &pid,
                      &error))
