@@ -50,7 +50,6 @@
 
 #include <pwd.h>
 #include <grp.h>
-#include <security/pam_appl.h>
 
 #include <xf86drm.h>
 
@@ -68,8 +67,6 @@ enum vt_state {
 };
 
 struct weston_launch {
-	struct pam_conv pc;
-	pam_handle_t *ph;
 	int tty;
 	int ttynr;
 	int sock[2];
@@ -79,7 +76,6 @@ struct weston_launch {
 
 	pid_t child;
 	int verbose;
-	char *new_user;
 
 	struct termios terminal_attributes;
 	int kb_mode;
@@ -218,7 +214,7 @@ handle_setdrmfd(struct weston_launch *wl, struct msghdr *msg, ssize_t len)
 		reply.ret = -errno;
 
 	if (wl->verbose)
-		fprintf(stderr, "weston-launch: set drm FD, ret: %d, fd: %d\n",
+		fprintf(stderr, "mutter-launch: set drm FD, ret: %d, fd: %d\n",
 			reply.ret, data->fd);
 
 out:
@@ -259,7 +255,7 @@ handle_confirm_vt_switch(struct weston_launch *wl, struct msghdr *msg, ssize_t l
 	ioctl(wl->tty, VT_RELDISP, 1);
 
 	if (wl->verbose)
-		fprintf(stderr, "weston-launcher: confirmed VT switch\n");
+		fprintf(stderr, "mutter-launcher: confirmed VT switch\n");
 
 	reply.ret = 0;
 
@@ -294,7 +290,7 @@ handle_activate_vt(struct weston_launch *wl, struct msghdr *msg, ssize_t len)
 		reply.ret = -errno;
 
 	if (wl->verbose)
-		fprintf(stderr, "weston-launch: activate VT, ret: %d\n", reply.ret);
+		fprintf(stderr, "mutter-launch: activate VT, ret: %d\n", reply.ret);
 
 out:
 	do {
@@ -372,7 +368,7 @@ err0:
 	iov.iov_len = sizeof reply;
 
 	if (wl->verbose)
-		fprintf(stderr, "weston-launch: opened %s: ret: %d, fd: %d\n",
+		fprintf(stderr, "mutter-launch: opened %s: ret: %d, fd: %d\n",
 			message->path, reply.ret, fd);
 	do {
 		len = sendmsg(wl->sock[0], &nmsg, 0);
@@ -453,8 +449,6 @@ tty_reset(struct weston_launch *wl)
 static void
 quit(struct weston_launch *wl, int status)
 {
-	int err;
-
 	if (wl->child > 0)
 		kill(wl->child, SIGKILL);
 
@@ -466,14 +460,6 @@ quit(struct weston_launch *wl, int status)
 
 	tty_reset(wl);
 
-	if (wl->new_user) {
-		err = pam_close_session(wl->ph, 0);
-		if (err)
-			fprintf(stderr, "pam_close_session failed: %d: %s\n",
-				err, pam_strerror(wl->ph, err));
-		pam_end(wl->ph, err);
-	}
-
 	exit(status);
 }
 
@@ -578,11 +564,24 @@ setup_tty(struct weston_launch *wl)
 
 	ok = sd_session_get_tty(session, &tty);
 	if (ok == 0) {
-		snprintf(path, PATH_MAX, "/dev/%s", tty);
+		/* Old systemd only has the tty name in the TTY
+		   field, new one has the full char device path.
+
+		   Check what we have and fix it properly.
+		*/
+		if (strncmp(tty, "/dev", strlen("/dev")) == 0) {
+			strncpy(path, tty, PATH_MAX);
+			path[PATH_MAX-1] = 0;
+		} else {
+			snprintf(path, PATH_MAX, "/dev/%s", tty);
+		}
+
 		wl->tty = open(path, O_RDWR | O_NOCTTY | O_CLOEXEC);
 		free(tty);
 #ifdef HAVE_SD_SESSION_GET_VT
 	} else if (ok == -ENOENT) {
+		unsigned vt;
+
 		/* Negative errnos are cool, right?
 		   So cool that we can't distinguish "session not found"
 		   from "key does not exist in the session file"!
@@ -590,11 +589,11 @@ setup_tty(struct weston_launch *wl)
 		   from sd_pid_get_session()...
 		*/
 
-		ok = sd_session_get_vt(session, &tty);
+		ok = sd_session_get_vt(session, &vt);
 		if (ok < 0)
 			error(1, -ok, "could not determine current TTY");
 
-		snprintf(path, PATH_MAX, "/dev/tty%s", tty);
+		snprintf(path, PATH_MAX, "/dev/tty%u", vt);
 		wl->tty = open(path, O_RDWR | O_NOCTTY | O_CLOEXEC);
 		free(tty);
 #endif
@@ -729,7 +728,7 @@ main(int argc, char *argv[])
 			wl.verbose = 1;
 			break;
 		case 'h':
-			help("weston-launch");
+			help("mutter-launch");
 			exit(EXIT_FAILURE);
 		}
 	}
